

#include <nrf_i2s.h>
#include <assert.h>


//I2S Pins
#define SCK_PIN 29      //Serial clock for i2s
#define MCK_PIN 30      //Master clock pin optional 
#define LRCK_PIN 44     //Word select clock
#define SDOUT_PIN 45    //Data Output
#define NC 0xFFFFFFFF   //Not connected

//sine wave parameters
#define SAMPLERATE_HZ       44100               // I2S Audio sample rate (8-96kHz range). Max98357 i2s amp - PWM is at 330kHz. 50kHz works well. 96kHz may be too fast.
#define CARRIER_FREQ        2000               // Sine wave carrier freq
#define MAX_AMPLITUDE       0.15*pow(2,16)-1    // C*pow(2,31)-1 is max. Sets the amplitude of generated waveforms, C value prevents saturation/distortion.
#define STIM_FREQ           25                  //Burst frequecny of waves
#define SINE_CYCLES         10                  //Sine wave cycles needed in single burst
#define SAMPLES_PER_CYCLE   int(SAMPLERATE_HZ / CARRIER_FREQ) //Samples per cycle for configured sine cycles with carrier freq
#define WAVE_SIZE           2*SAMPLES_PER_CYCLE*SINE_CYCLES   // Must be EVEN (since interleaving L and R channels) so 2*SAMPLES_PER_CYCLE*SINE_CYCLES. defines size of each generated waveform (one cycle).  The larger the size the higher the quality.


//buffer for sine wave data samples
int16_t Wave_data[WAVE_SIZE] = {0};
//buffer for dma transmission
int16_t Wave_DMA[WAVE_SIZE] = {0};

//DMA transmission count status flag
static uint8_t flag = 0;

//I2S DMA interrupt handler
void i2s_irq_handler();


//function prototypes
void allocate_ram_address();
void wave_generate();
void Wave_trasnfer();
void I2S_init();
void I2S_Handler_allocation();

void setup()
{
  //Serial monitor initializing
  Serial.begin(9600);

  //populate sine wave data for the configured WAVE_SIZE
  wave_generate();

  //populate Wave_DMA buffer with Wave_data buffer values multiplied with certain amplitude
  Wave_trasnfer();

  //initialize I2S
  I2S_init();

  //configure i2s interrupt handler
  I2S_Handler_allocation();

  //checking buffer is in RAM
  allocate_ram_address();

  //start transfer of i2s data
  nrf_i2s_task_trigger(NRF_I2S, NRF_I2S_TASK_START);
}


// the loop function runs over and over again forever
void loop()
{
  //restart the trasnmission.
  if (flag == 1)
  {
    //start transfer of i2s data
    nrf_i2s_task_trigger(NRF_I2S, NRF_I2S_TASK_START);
  }
  delay(5000);
}

//helper functions

/*
   @brief Function for setting up I2S Pheripheral.

   This function is used to sets up i2s sck , lrck , data output pins as per user configuration.
   I2S is configured as master with stereo mode and sampling frequency is set at 44100Hz
   Sampling data width is set to 16 bit and data is left aligned, typical I2S format is used.
   I2S TXD pointer is assigned to start address of Wave_DMA buffer and I2S Pheripheral and
   trasnmission event interrupt is enabled.

   if the i2s configuration is valid ,i2s configuration successfull message will be printed.

*/
void I2S_init()
{
  //initializing i2s pins
  nrf_i2s_pins_set(NRF_I2S, SCK_PIN, LRCK_PIN, NC, SDOUT_PIN, NC);

  //configuring i2s following parameters
  bool i2s_config_status = false;

  /*
     I2S MASTER CONFIGURATION
     MCK    10.6666667 MHz  (32MHz/3)
     LRCK   44100 Hz        (MCK/256)
     SCK    1.4MHz (2*Sampling_rate*Sample_width)
     Sample Width 16 Bit
     Left aligned sample data
     Stereo channel mode
     I2S frame format
  */
  i2s_config_status = nrf_i2s_configure(NRF_I2S,                            //I2S instance
                                        NRF_I2S_MODE_MASTER,                //I2S configured as master
                                        NRF_I2S_FORMAT_I2S,                 //I2S frame format
                                        NRF_I2S_ALIGN_LEFT,                 //Alignment of sample within a frame
                                        NRF_I2S_SWIDTH_16BIT,               //Sample width
                                        NRF_I2S_CHANNELS_STEREO,            //Enabled channels
                                        NRF_I2S_MCK_32MDIV3,                //Master clock generator setup
                                        NRF_I2S_RATIO_256X                  //MCK/LRCK ratio

                                       );
  //checking status of configuration
  i2s_config_status ? Serial.println("i2s configuration successfull") : Serial.println("i2s configuration failed");

  //setting up the I2S transfer
  nrf_i2s_transfer_set(NRF_I2S, WAVE_SIZE, NULL, (uint32_t const *)Wave_DMA);

  //enable i2s peripheral
  nrf_i2s_enable(NRF_I2S);

  Serial.println("nrf_i2s_enabled");

  //enable i2s tx event interrupt
  nrf_i2s_int_enable(NRF_I2S, NRF_I2S_INT_TXPTRUPD_MASK);

}


/**
   @brief Function for checking if an object is placed in the Data RAM region.

   Several peripherals (the ones using EasyDMA) require the transfer buffers
   to be placed in the Data RAM region. This function can be used to check if
   this condition is met.
**/
void allocate_ram_address()
{
  assert(Wave_DMA != NULL);
  assert(nrfx_is_in_ram(Wave_DMA) && is_word_aligned(Wave_DMA));
}

/*
   @brief Function used for setting i2s interrupt handler using vector table

   i2s interrupt handler is set to i2s_irq_handler using NVIC and vector table
   i2s inetrrupt is enbaled using NVIC registers and priority is set to NRFX_I2S_CONFIG_IRQ_PRIORITY

   if i2s interrupt is enabled NVIC i2s enabled will be printed
*/
void I2S_Handler_allocation()
{
  //assigning i2s_irq_handler as i2s interrupt handler
  __NVIC_SetVector(I2S_IRQn, (uint32_t)i2s_irq_handler);

  //setting I2S interrupt priority
  NRFX_IRQ_PRIORITY_SET(I2S_IRQn, NRFX_I2S_CONFIG_IRQ_PRIORITY);

  //enabling I2S interrupt in NVIC
  NRFX_IRQ_ENABLE(I2S_IRQn);

  //I2S interrupt status
  __NVIC_GetEnableIRQ(I2S_IRQn) ? Serial.println("NVIC i2s enabled") : Serial.println("NVIC i2s disabled");

}

/*
    @brief This function used for generating i2s wave samples for the
    configured WAVE_SIZE.

    Left and Right channels of i2s is filled using sine wave data sample with
    specified carrier frequency.
*/
void wave_generate()
{
  int16_t *ptr = Wave_data;
  // Populate buffer values with waveform:
  // the I2S module (MAX98357) will be expecting data interleaved LRLR (unless SD input pin is used):
  for (uint32_t i = 0; i < WAVE_SIZE / 2; i++)
  {
    /* create a sine wave on the left channel */
    *ptr++ = int32_t(float((MAX_AMPLITUDE) * sin((2.0 * SINE_CYCLES * PI * (1.0 / (WAVE_SIZE / 2))) * i)));
    /* create a sine wave on the right channel */
    *ptr++ = int32_t(float((MAX_AMPLITUDE) * sin((2.0 * SINE_CYCLES * PI * (1.0 / (WAVE_SIZE / 2))) * i)));
  }
}


/*
   @brief This function is used to set amplitude of sine wave data samples.

   Sine wave data samples are multiplied with data recieved from Blynk App
   Current amplitude is set to default 100% for all samples.
*/

void Wave_trasnfer()
{

  for (int i = 0; i < WAVE_SIZE; ++i)
  { // V10_value=stim amplitude in % (100% is max)
    Wave_DMA[i] = int32_t(float(100 * 100 / 100.0) * Wave_data[i]);
  }  // scaling MUCH faster (<5ms) than calling setupSine (~60ms) to rebuild waveform

}

/*
   @brief This is the interrupt handler for i2s peripheral

   once the Wave_DMA buffer is copied to Hardware buffer of i2s ,TXPTRUPD event is generated

   initial TXPTRUPD event is ignored since Wave_DMA buffer is copied for the first time to buffer.
   the next TXPTRUPD event is generated when Hardware buffer is updated for the second time
   Second TXPTRUPD event is occured once initial transmission is completed , during the TXPTRUPD
   events a flag is incremented and when the flag is set to 2 , I2S transmission is stopped and
   flag reset to 0.
*/
void i2s_irq_handler(void)
{
  //Checking TXPTRUPD event
  if (nrf_i2s_event_check(NRF_I2S, NRF_I2S_EVENT_TXPTRUPD))
  {
    //clear TXPTRUPD event
    nrf_i2s_event_clear(NRF_I2S, NRF_I2S_EVENT_TXPTRUPD);
    //increment transmission complete flag
    flag++;
    if (flag > 2)
    {
      //stopping i2s transfer
      nrf_i2s_task_trigger(NRF_I2S, NRF_I2S_TASK_STOP);
      //set the transmission complete flag to 0.
      flag = 0;
    }
  }
}